@@ -4,76 +4,36 @@
 //! time and inverting them.
 
 use std::convert::TryInto;
-use std::io::{BufRead, BufReader, Read, Write};
+use std::io::{stdin, stdout, BufRead, BufReader, Read, Write};
 use std::sync::mpsc::{channel, Receiver, Sender, TryRecvError};
-
 use std::time::SystemTime;
 
 use anyhow::Context;
-
 use cursive::event::Key;
-use cursive::theme::{Color, PaletteColor};
-use cursive::traits::Nameable;
-use cursive::views::{
-    Dialog, EditView, LinearLayout, NamedView, OnEventView, ScrollView, TextView,
-};
+use cursive::utils::markup::StyledString;
+use cursive::views::{Dialog, EditView, LinearLayout, OnEventView, ScrollView, TextView};
 use cursive::{Cursive, CursiveRunnable, CursiveRunner};
 
 use crate::commands::smartlog::render_graph;
 use crate::core::eventlog::{Event, EventCursor, EventLogDb, EventReplayer, EventTransactionId};
-use crate::core::formatting::{Glyphs, Pluralize};
+use crate::core::formatting::{printable_styled_string, Glyphs, Pluralize, StyledStringBuilder};
 use crate::core::graph::{make_graph, BranchOids, HeadOid, MainBranchOid};
 use crate::core::mergebase::MergeBaseDb;
 use crate::core::metadata::{
     render_commit_metadata, BranchesProvider, CommitMessageProvider, CommitOidProvider,
     DifferentialRevisionProvider, HiddenExplanationProvider, RelativeTimeProvider,
 };
+use crate::core::tui::{with_siv, SingletonView};
+use crate::declare_views;
 use crate::util::{get_db_conn, get_repo, run_git, GitExecutable};
 
-pub(crate) fn with_siv<T, F: FnOnce(CursiveRunner<CursiveRunnable>) -> anyhow::Result<T>>(
-    f: F,
-) -> anyhow::Result<T> {
-    // I tried these back-ends:
-    //
-    // * `ncurses`/`pancurses`: Doesn't render ANSI escape codes. (NB: the fact
-    //   that we print out strings with ANSI escape codes is tech debt; we would
-    //   ideally pass around styled representations of all text, and decide how to
-    //   rendering it later.) Rendered scroll view improperly. No mouse/scrolling
-    //   support
-    // * `termion`: Renders ANSI escape codes. Has mouse/scrolling support. But
-    //   critical bug: https://github.com/gyscos/cursive/issues/563
-    // * `crossterm`: Renders ANSI escape codes. Has mouse/scrolling support.
-    //   However, has some flickering issues, particularly when scrolling. See
-    //   issue at https://github.com/gyscos/cursive/issues/142. I tried the
-    //   `cursive_buffered_backend` library, but this causes it to no longer
-    //   respect the ANSI escape codes.
-    // * `blt`: Seems to require that a certain library be present on the system
-    //   for linking.
-    let mut siv = cursive::crossterm();
-
-    siv.update_theme(|theme| {
-        theme.shadow = false;
-        theme.palette.extend(vec![
-            (PaletteColor::Background, Color::TerminalDefault),
-            (PaletteColor::View, Color::TerminalDefault),
-            (PaletteColor::Primary, Color::TerminalDefault),
-        ]);
-    });
-    let old_max_level = log::max_level();
-    log::set_max_level(log::LevelFilter::Off);
-    let result = f(siv.into_runner());
-    log::set_max_level(old_max_level);
-    let result = result?;
-    Ok(result)
-}
-
 fn render_cursor_smartlog(
     glyphs: &Glyphs,
     repo: &git2::Repository,
     merge_base_db: &MergeBaseDb,
     event_replayer: &EventReplayer,
     event_cursor: EventCursor,
-) -> anyhow::Result<String> {
+) -> anyhow::Result<Vec<StyledString>> {
     let head_oid = event_replayer.get_cursor_head_oid(event_cursor);
     let main_branch_oid = event_replayer.get_cursor_main_branch_oid(event_cursor, repo)?;
     let branch_oid_to_names = event_replayer.get_cursor_branch_oid_to_names(event_cursor, repo)?;
@@ -87,24 +47,21 @@ fn render_cursor_smartlog(
         &BranchOids(branch_oid_to_names.keys().copied().collect()),
         true,
     )?;
-    let mut out = Vec::new();
-    render_graph(
-        &mut out,
+    let result = render_graph(
         glyphs,
         repo,
         merge_base_db,
         &graph,
         &HeadOid(head_oid),
-        &[
-            &CommitOidProvider::new(true)?,
-            &RelativeTimeProvider::new(&repo, SystemTime::now())?,
-            &HiddenExplanationProvider::new(&graph, &event_replayer, event_cursor)?,
-            &BranchesProvider::new(&repo, &branch_oid_to_names)?,
-            &DifferentialRevisionProvider::new(&repo)?,
-            &CommitMessageProvider::new()?,
+        &mut [
+            &mut CommitOidProvider::new(true)?,
+            &mut RelativeTimeProvider::new(&repo, SystemTime::now())?,
+            &mut HiddenExplanationProvider::new(&graph, &event_replayer, event_cursor)?,
+            &mut BranchesProvider::new(&repo, &branch_oid_to_names)?,
+            &mut DifferentialRevisionProvider::new(&repo)?,
+            &mut CommitMessageProvider::new()?,
         ],
     )?;
-    let result = String::from_utf8(out)?;
     Ok(result)
 }
 
@@ -115,20 +72,20 @@ fn render_ref_name(ref_name: &str) -> String {
     }
 }
 
-fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<String> {
-    let render_commit = |oid: git2::Oid| -> anyhow::Result<String> {
+fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Vec<StyledString>> {
+    let render_commit = |oid: git2::Oid| -> anyhow::Result<StyledString> {
         match repo.find_commit(oid) {
             Ok(commit) => render_commit_metadata(
                 &commit,
-                &[
-                    &CommitOidProvider::new(true)?,
-                    &CommitMessageProvider::new()?,
+                &mut [
+                    &mut CommitOidProvider::new(true)?,
+                    &mut CommitMessageProvider::new()?,
                 ],
             ),
-            Err(_) => Ok(format!(
+            Err(_) => Ok(StyledString::plain(format!(
                 "<unavailable: {} (possibly GC'ed)>",
                 oid.to_string()
-            )),
+            ))),
         }
     };
     let result = match event {
@@ -137,7 +94,13 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Stri
             event_tx_id: _,
             commit_oid,
         } => {
-            format!("Commit {}\n", render_commit(*commit_oid)?)
+            vec![
+                StyledStringBuilder::new()
+                    .append_plain("Commit ")
+                    .append(render_commit(*commit_oid)?)
+                    .build(),
+                StyledString::new(),
+            ]
         }
 
         Event::HideEvent {
@@ -145,7 +108,13 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Stri
             event_tx_id: _,
             commit_oid,
         } => {
-            format!("Hide commit {}\n", render_commit(*commit_oid)?)
+            vec![
+                StyledStringBuilder::new()
+                    .append_plain("Hide commit ")
+                    .append(render_commit(*commit_oid)?)
+                    .build(),
+                StyledString::new(),
+            ]
         }
 
         Event::UnhideEvent {
@@ -153,7 +122,13 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Stri
             event_tx_id: _,
             commit_oid,
         } => {
-            format!("Unhide commit {}\n", render_commit(*commit_oid)?)
+            vec![
+                StyledStringBuilder::new()
+                    .append_plain("Unhide commit ")
+                    .append(render_commit(*commit_oid)?)
+                    .build(),
+                StyledString::new(),
+            ]
         }
 
         Event::RefUpdateEvent {
@@ -165,7 +140,13 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Stri
             message: _,
         } if ref_name == "HEAD" => {
             // Not sure if this can happen. When a repo is created, maybe?
-            format!("Check out to {}\n", render_commit(new_ref.parse()?)?)
+            vec![
+                StyledStringBuilder::new()
+                    .append_plain("Check out to ")
+                    .append(render_commit(new_ref.parse()?)?)
+                    .build(),
+                StyledString::new(),
+            ]
         }
 
         Event::RefUpdateEvent {
@@ -176,13 +157,16 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Stri
             new_ref: Some(new_ref),
             message: _,
         } if ref_name == "HEAD" => {
-            format!(
-                "\
-Check out from {}
-            to {}",
-                render_commit(old_ref.parse()?)?,
-                render_commit(new_ref.parse()?)?
-            )
+            vec![
+                StyledStringBuilder::new()
+                    .append_plain("Check out from ")
+                    .append(render_commit(old_ref.parse()?)?)
+                    .build(),
+                StyledStringBuilder::new()
+                    .append_plain("            to ")
+                    .append(render_commit(new_ref.parse()?)?)
+                    .build(),
+            ]
         }
 
         Event::RefUpdateEvent {
@@ -193,13 +177,17 @@ Check out from {}
             new_ref: None,
             message: _,
         } => {
-            format!(
-                "\
-Empty event for {}
-This event should not appear. This is a (benign) bug -- please report it.
-",
-                render_ref_name(ref_name)
-            )
+            vec![
+                StyledStringBuilder::new()
+                    .append_plain("Empty event for ")
+                    .append_plain(render_ref_name(ref_name))
+                    .build(),
+                StyledStringBuilder::new()
+                    .append_plain("This event should not appear. ")
+                    .append_plain("This is a (benign) bug -- ")
+                    .append_plain("please report it.")
+                    .build(),
+            ]
         }
 
         Event::RefUpdateEvent {
@@ -210,11 +198,15 @@ This event should not appear. This is a (benign) bug -- please report it.
             new_ref: Some(new_ref),
             message: _,
         } => {
-            format!(
-                "Create {} at {}\n",
-                render_ref_name(ref_name),
-                render_commit(new_ref.parse()?)?
-            )
+            vec![
+                StyledStringBuilder::new()
+                    .append_plain("Create ")
+                    .append_plain(render_ref_name(ref_name))
+                    .append_plain(" at ")
+                    .append(render_commit(new_ref.parse()?)?)
+                    .build(),
+                StyledString::new(),
+            ]
         }
 
         Event::RefUpdateEvent {
@@ -225,11 +217,15 @@ This event should not appear. This is a (benign) bug -- please report it.
             new_ref: None,
             message: _,
         } => {
-            format!(
-                "Delete {} at {}\n",
-                render_ref_name(ref_name),
-                render_commit(old_ref.parse()?)?
-            )
+            vec![
+                StyledStringBuilder::new()
+                    .append_plain("Delete ")
+                    .append_plain(render_ref_name(ref_name))
+                    .append_plain(" at ")
+                    .append(render_commit(old_ref.parse()?)?)
+                    .build(),
+                StyledString::new(),
+            ]
         }
 
         Event::RefUpdateEvent {
@@ -241,15 +237,20 @@ This event should not appear. This is a (benign) bug -- please report it.
             message: _,
         } => {
             let ref_name = render_ref_name(ref_name);
-            format!(
-                "\
-Move {} from {}
-     {}   to {}",
-                ref_name,
-                render_commit(old_ref.parse()?)?,
-                " ".repeat(ref_name.len()),
-                render_commit(new_ref.parse()?)?,
-            )
+            vec![
+                StyledStringBuilder::new()
+                    .append_plain("Move ")
+                    .append_plain(ref_name.clone())
+                    .append_plain(" from ")
+                    .append(render_commit(old_ref.parse()?)?)
+                    .build(),
+                StyledStringBuilder::new()
+                    .append_plain("     ")
+                    .append_plain(" ".repeat(ref_name.len()))
+                    .append_plain("   to ")
+                    .append(render_commit(new_ref.parse()?)?)
+                    .build(),
+            ]
         }
 
         Event::RewriteEvent {
@@ -258,35 +259,43 @@ Move {} from {}
             old_commit_oid,
             new_commit_oid,
         } => {
-            format!(
-                "\
-Rewrite commit {}
-            as {}",
-                render_commit(*old_commit_oid)?,
-                render_commit(*new_commit_oid)?
-            )
+            vec![
+                StyledStringBuilder::new()
+                    .append_plain("Rewrite commit ")
+                    .append(render_commit(*old_commit_oid)?)
+                    .build(),
+                StyledStringBuilder::new()
+                    .append_plain("           as ")
+                    .append(render_commit(*new_commit_oid)?)
+                    .build(),
+            ]
         }
     };
     Ok(result)
 }
 
 fn describe_events_numbered(
-    out: &mut impl Write,
     repo: &git2::Repository,
     events: &[Event],
-) -> Result<(), anyhow::Error> {
+) -> Result<Vec<StyledString>, anyhow::Error> {
+    let mut lines = Vec::new();
     for (i, event) in (1..).zip(events) {
         let num_header = format!("{}. ", i);
-        for (j, line) in (0..).zip(describe_event(&repo, &event)?.split('\n')) {
-            if j == 0 {
-                write!(out, "{}", num_header)?;
+        for (j, event_line) in (0..).zip(describe_event(&repo, &event)?) {
+            let prefix = if j == 0 {
+                num_header.clone()
             } else {
-                write!(out, "{}", " ".repeat(num_header.len()))?;
-            }
-            writeln!(out, "{}", line)?;
+                " ".repeat(num_header.len())
+            };
+            lines.push(
+                StyledStringBuilder::new()
+                    .append_plain(prefix)
+                    .append(event_line)
+                    .build(),
+            );
         }
     }
-    Ok(())
+    Ok(lines)
 }
 
 fn select_past_event(
@@ -350,10 +359,11 @@ fn select_past_event(
             siv.step();
         }
 
-        type SmartlogView = ScrollView<TextView>;
-        const SMARTLOG_VIEW_NAME: &str = "smartlog-view";
-        type InfoView = TextView;
-        const INFO_VIEW_NAME: &str = "info-view";
+        declare_views! {
+            SmartlogView => ScrollView<TextView>,
+            InfoView => TextView,
+        }
+
         let redraw = |siv: &mut Cursive,
                       event_replayer: &mut EventReplayer,
                       event_cursor: EventCursor|
@@ -365,19 +375,19 @@ fn select_past_event(
                 &event_replayer,
                 event_cursor,
             )?;
-            siv.find_name::<SmartlogView>(SMARTLOG_VIEW_NAME)
-                .unwrap()
+            SmartlogView::find(siv)
                 .get_inner_mut()
-                .set_content(smartlog);
+                .set_content(StyledStringBuilder::from_lines(smartlog));
 
             let event = event_replayer.get_tx_events_before_cursor(event_cursor);
             let info_view_contents = match event {
-                None => "There are no previous available events.".to_owned(),
+                None => vec![StyledString::plain(
+                    "There are no previous available events.",
+                )],
                 Some((event_id, events)) => {
                     let event_description = {
-                        let mut out = Vec::new();
-                        describe_events_numbered(&mut out, repo, &events)?;
-                        String::from_utf8(out)?
+                        let lines = describe_events_numbered(repo, &events)?;
+                        StyledStringBuilder::from_lines(lines)
                     };
                     let relative_time_provider = RelativeTimeProvider::new(repo, now)?;
                     let relative_time = if relative_time_provider.is_enabled() {
@@ -391,18 +401,21 @@ fn select_past_event(
                     } else {
                         String::new()
                     };
-                    format!(
-                            "Repo after transaction {event_tx_id} (event {event_id}){relative_time}. Press 'h' for help, 'q' to quit.\n{event_description}\n",
-                            event_tx_id = events[0].get_event_tx_id().to_string(),
-                            event_id = event_id,
-                            relative_time = relative_time,
-                            event_description = event_description,
-                        )
+                    vec![
+                        StyledStringBuilder::new()
+                            .append_plain("Repo after transaction ")
+                            .append_plain(events[0].get_event_tx_id().to_string())
+                            .append_plain(" (event ")
+                            .append_plain(event_id.to_string())
+                            .append_plain(")")
+                            .append_plain(relative_time)
+                            .append_plain(". Press 'h' for help, 'q' to quit.")
+                            .build(),
+                        event_description,
+                    ]
                 }
             };
-            siv.find_name::<InfoView>(INFO_VIEW_NAME)
-                .unwrap()
-                .set_content(info_view_contents);
+            InfoView::find(siv).set_content(StyledStringBuilder::from_lines(info_view_contents));
             Ok(())
         };
 
@@ -416,9 +429,8 @@ fn select_past_event(
             }
 
             Ok(Message::Init) => {
-                let smartlog_view: NamedView<SmartlogView> =
-                    ScrollView::new(TextView::new("")).with_name(SMARTLOG_VIEW_NAME);
-                let info_view: NamedView<InfoView> = TextView::new("").with_name(INFO_VIEW_NAME);
+                let smartlog_view: SmartlogView = ScrollView::new(TextView::new("")).into();
+                let info_view: InfoView = TextView::new("").into();
                 siv.add_layer(
                     LinearLayout::vertical()
                         .child(smartlog_view)
@@ -496,10 +508,7 @@ You can also copy a commit hash from the past and manually run `git unhide` or `
 
             Ok(Message::SelectEventIdAndQuit) => {
                 siv.quit();
-                match event_replayer.get_event_before_cursor(cursor) {
-                    Some(_) => return Ok(Some(cursor)),
-                    None => return Ok(None),
-                }
+                return Ok(Some(cursor));
             }
         };
 
@@ -597,7 +606,7 @@ fn optimize_inverse_events(events: Vec<Event>) -> Vec<Event> {
 fn undo_events(
     in_: &mut impl Read,
     out: &mut impl Write,
-    err: &mut impl Write,
+    glyphs: &Glyphs,
     repo: &git2::Repository,
     git_executable: &GitExecutable,
     event_log_db: &mut EventLogDb,
@@ -625,14 +634,27 @@ fn undo_events(
         })
         .map(|event| inverse_event(event.clone(), now, event_tx_id))
         .collect::<anyhow::Result<Vec<Event>>>()?;
-    let inverse_events = optimize_inverse_events(inverse_events);
+    let mut inverse_events = optimize_inverse_events(inverse_events);
+
+    // Move any checkout operations to be first. Otherwise, we have the risk
+    // that `HEAD` is a symbolic reference pointing to another reference, and we
+    // update that reference. This would cause the working copy to become dirty
+    // from Git's perspective.
+    inverse_events.sort_by_key(|event| match event {
+        Event::RefUpdateEvent { ref_name, .. } if ref_name == "HEAD" => 0,
+        _ => 1,
+    });
+
     if inverse_events.is_empty() {
         writeln!(out, "No undo actions to apply, exiting.")?;
         return Ok(0);
     }
 
     writeln!(out, "Will apply these actions:")?;
-    describe_events_numbered(out, repo, &inverse_events)?;
+    let events = describe_events_numbered(&repo, &inverse_events)?;
+    for line in events {
+        writeln!(out, "{}", printable_styled_string(&glyphs, line)?)?;
+    }
 
     let confirmed = {
         write!(out, "Confirm? [yN] ")?;
@@ -658,6 +680,7 @@ fn undo_events(
         plural: "inverse events",
     }
     .to_string();
+
     for event in inverse_events.into_iter() {
         match event {
             Event::RefUpdateEvent {
@@ -673,11 +696,9 @@ fn undo_events(
                 // dirty working copy). The `Git` command will update the event
                 // log appropriately, as it will invoke our hooks.
                 run_git(
-                    out,
-                    err,
                     git_executable,
                     Some(event_tx_id),
-                    &["checkout", &new_ref],
+                    &["checkout", "--detach", &new_ref],
                 )
                 .with_context(|| "Updating to previous HEAD location")?;
             }
@@ -746,12 +767,7 @@ fn undo_events(
 }
 
 /// Restore the repository to a previous state interactively.
-pub fn undo(
-    in_: &mut impl Read,
-    out: &mut impl Write,
-    err: &mut impl Write,
-    git_executable: &GitExecutable,
-) -> anyhow::Result<isize> {
+pub fn undo(git_executable: &GitExecutable) -> anyhow::Result<isize> {
     let glyphs = Glyphs::detect();
     let repo = get_repo()?;
     let conn = get_db_conn(&repo)?;
@@ -770,9 +786,9 @@ pub fn undo(
     };
 
     let result = undo_events(
-        in_,
-        out,
-        err,
+        &mut stdin(),
+        &mut stdout().lock(),
+        &glyphs,
         &repo,
         &git_executable,
         &mut event_log_db,
@@ -793,12 +809,6 @@ pub mod testing {
     use crate::core::mergebase::MergeBaseDb;
     use crate::util::GitExecutable;
 
-    pub fn with_siv<T, F: FnOnce(CursiveRunner<CursiveRunnable>) -> anyhow::Result<T>>(
-        f: F,
-    ) -> anyhow::Result<T> {
-        super::with_siv(f)
-    }
-
     pub fn select_past_event(
         siv: CursiveRunner<CursiveRunnable>,
         glyphs: &Glyphs,
@@ -812,7 +822,7 @@ pub mod testing {
     pub fn undo_events(
         in_: &mut impl Read,
         out: &mut impl Write,
-        err: &mut impl Write,
+        glyphs: &Glyphs,
         repo: &git2::Repository,
         git_executable: &GitExecutable,
         event_log_db: &mut EventLogDb,
@@ -822,7 +832,7 @@ pub mod testing {
         super::undo_events(
             in_,
             out,
-            err,
+            glyphs,
             repo,
             git_executable,
             event_log_db,
