@@ -7,7 +7,7 @@ use std::path::{Path, PathBuf};
 use std::process::Command;
 use std::str::FromStr;
 
-use crate::util::{wrap_git_error, GitExecutable, GitVersion};
+use crate::util::{get_sh, wrap_git_error, GitExecutable, GitVersion};
 use anyhow::Context;
 use fn_error_context::context;
 
@@ -16,16 +16,16 @@ const DUMMY_EMAIL: &str = "test@example.com";
 const DUMMY_DATE: &str = "Wed 29 Oct 12:34:56 2020 PDT";
 
 /// Wrapper around the Git executable, for testing.
-#[derive(Debug)]
-pub struct Git<'a> {
+#[derive(Clone, Debug)]
+pub struct Git {
     /// The path to the repository on disk. The directory itself must exist,
     /// although it might not have a `.git` folder in it. (Use `Git::init_repo`
     /// to initialize it.)
-    pub repo_path: &'a Path,
+    pub repo_path: PathBuf,
 
     /// The path to the Git executable on disk. This is important since we test
     /// against multiple Git versions.
-    pub git_executable: &'a Path,
+    pub git_executable: PathBuf,
 }
 
 /// Options for `Git::init_repo_with_options`.
@@ -69,9 +69,9 @@ impl Default for GitRunOptions {
     }
 }
 
-impl<'a> Git<'a> {
+impl Git {
     /// Constructor.
-    pub fn new(repo_path: &'a Path, git_executable: &'a GitExecutable) -> Self {
+    pub fn new(repo_path: PathBuf, git_executable: GitExecutable) -> Self {
         let GitExecutable(git_executable) = git_executable;
         Git {
             repo_path,
@@ -99,16 +99,23 @@ impl<'a> Git<'a> {
             .git_executable
             .parent()
             .expect("Unable to find git path parent");
-        vec![
-            // For Git to be able to launch `git-branchless`.
-            branchless_path,
-            // For our hooks to be able to call back into `git`.
-            git_path,
-        ]
-        .iter()
-        .map(|path| path.to_str().expect("Unable to decode path component"))
-        .collect::<Vec<_>>()
-        .join(":")
+        let bash = get_sh().expect("bash missing?");
+        let bash_path = bash.parent().unwrap();
+        std::env::join_paths(
+            vec![
+                // For Git to be able to launch `git-branchless`.
+                branchless_path,
+                // For our hooks to be able to call back into `git`.
+                git_path,
+                // For branchless to manually invoke bash when needed.
+                bash_path,
+            ]
+            .iter()
+            .map(|path| path.to_str().expect("Unable to decode path component")),
+        )
+        .expect("joining paths")
+        .to_string_lossy()
+        .to_string()
     }
 
     /// Run a Git command.
@@ -124,11 +131,16 @@ impl<'a> Git<'a> {
             expected_exit_code,
         } = options;
 
-        let git_executable = if !use_system_git {
-            self.git_executable
+        let default_path = if cfg!(target_os = "windows") {
+            Path::new("C:\\Program Files\\Git\\bin\\git.exe")
         } else {
             Path::new("/usr/bin/git")
         };
+        let git_executable = if !use_system_git {
+            &self.git_executable
+        } else {
+            default_path
+        };
 
         // Required for determinism, as these values will be baked into the commit
         // hash.
@@ -149,15 +161,9 @@ impl<'a> Git<'a> {
             // messages. Usually, we can set this with `git commit -m`, but we have
             // no such option for things such as `git rebase`, which may call `git
             // commit` later as a part of their execution.
-            ("GIT_EDITOR", {
-                // Some systems have `/bin/true` and others have
-                // `/usr/bin/true`. Pick whichever one we can find.
-                if Path::new("/bin/true").exists() {
-                    "/bin/true"
-                } else {
-                    "/usr/bin/true"
-                }
-            }),
+            //
+            // ":" is understood by `git` to skip editing.
+            ("GIT_EDITOR", ":"),
             (
                 "PATH_TO_GIT",
                 git_executable
@@ -227,9 +233,6 @@ impl<'a> Git<'a> {
             self.commit_file("initial", 0)?;
         }
 
-        // Silence some log-spam.
-        self.run(&["config", "advice.detachedHead", "false"])?;
-
         // Non-deterministic metadata (depends on current time).
         self.run(&["config", "branchless.commitMetadata.relativeTime", "false"])?;
 
@@ -244,6 +247,13 @@ impl<'a> Git<'a> {
         self.init_repo_with_options(&Default::default())
     }
 
+    /// Write the provided contents to the provided file in the repository root.
+    pub fn write_file(&self, name: &str, contents: &str) -> anyhow::Result<()> {
+        let file_path = self.repo_path.join(format!("{}.txt", name));
+        std::fs::write(&file_path, contents)?;
+        Ok(())
+    }
+
     /// Commit a file with default contents. The `time` argument is used to set
     /// the commit timestamp, which is factored into the commit hash.
     #[context(
@@ -258,8 +268,7 @@ impl<'a> Git<'a> {
         time: isize,
         contents: &str,
     ) -> anyhow::Result<git2::Oid> {
-        let file_path = self.repo_path.join(format!("{}.txt", name));
-        std::fs::write(&file_path, contents)?;
+        self.write_file(name, contents)?;
         self.run(&["add", "."])?;
         self.run_with_options(
             &["commit", "-m", &format!("create {}.txt", name)],
@@ -339,7 +348,7 @@ pub fn get_git_executable() -> anyhow::Result<PathBuf> {
 pub fn with_git(f: fn(Git) -> anyhow::Result<()>) -> anyhow::Result<()> {
     let repo_dir = tempfile::tempdir()?;
     let git_executable = get_git_executable()?;
-    let git_executable = GitExecutable(&git_executable);
-    let git = Git::new(Path::new(repo_dir.path()), &git_executable);
+    let git_executable = GitExecutable(git_executable);
+    let git = Git::new(repo_dir.path().to_path_buf(), git_executable);
     f(git)
 }
